//==============================================================================
// BAPLIB - The main global definitions for the CPhenix program - still needs
//		to be cleaned up.
//
// Version 0.1 - Work in Progress
//
// Major Contributers: Kurt Eckhardt 
// 		Robert V Mahoney Jr
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//==============================================================================

// This program assumes that the main files were compiled as C files
#ifndef _BAPLIB_H_
#define _BAPLIB_H_


typedef unsigned char byte;			//			// define some simple data types.
typedef unsigned short word;		
typedef unsigned long ulong;		// 
#include "wiring.h"


#ifdef __cplusplus
extern "C" {
#endif


// include support headers.  Need to figure out more later..
//#define BASICATOMPRO40 1
//#define BASICATOMPROARC32 1

#ifndef BASICATOMPRO28
#ifndef   BASICATOMPRO40
#ifndef     BASICATOMPROARC32
#warning "Processor not defined using atom pro28"
#define BASICATOMPRO28 1
#endif
#endif
#endif

#ifdef BASICATOMPRO28 
#define H8_FAMILY_3694	
#ifndef F_CPU
#define F_CPU 16000000L
#endif

#include "3694s.h"

#else	// ARC32 or Bap40
#define H8_FAMILY_3687	
#ifndef F_CPU
#define F_CPU 20000000L
#endif

#include "3687s.h"
#endif

// Some defines to help determine what BAP we are running on at run time...
#define H8_3694_BAP28 0		// index for which tables to use for this
#define H8_3687_BAP40 10		// Could generate indexes, but for now save away...
#define H8_3687_ARC32 11		// 


#include "regdefs.h"


#define u8	byte
#define u16	word
#define u32	ulong
#define _FALSE_		0
#define _TRUE_		1

#include "string.h"
#include "printf.h"
#include "ArduinoPrint.h"
#include "iopins.h"		// has our IO definitions defined.

//=============================================================================
//[CONSTANTS]
//=============================================================================
#ifdef BASICATOMPRO28 		// Use new defines... need to change elsewhere...
#define	WTIMERTICSPERMSMUL  	64	// BAP28 is 16mhz need a multiplyer and divider to make the conversion with /8192
#define WTIMERTICSPERMSDIV  	125 // 
#define USEINT_TIMERAV
typedef enum {_MASTER_INT_, TIMERAINT} INTS;
#else
#define	WTIMERTICSPERMSMUL  	16	// BAP40 is 20mhz need a multiplyer and divider to make the conversion with /8192
#define WTIMERTICSPERMSDIV  	25 // 
typedef enum {_MASTER_INT_, TIMERB1INT} INTS;
#endif

// Define Interrupts we will process
extern void enable(INTS ints=_MASTER_INT_);
extern void disable(INTS ints=_MASTER_INT_);

#ifdef BASICATOMPRO28
#define bap_initialize()	_bap_initialize(H8_3694_BAP28);
#endif
#ifdef BASICATOMPRO40
#define bap_initialize()	_bap_initialize(H8_3687_BAP40);
#endif
#ifdef BASICATOMPROARC32
#define bap_initialize()	_bap_initialize(H8_3687_ARC32);	// Tell system to use second port item table
#endif

extern void _bap_initialize(u8 bWhichBap);

extern void InitSystemTimer(void);
extern unsigned long GetCurrentTime(void);
extern bool FTimeHasElapsed(u32 lTimeSave, u32 lTimeInMS);


#define HighByte(_wValue) ((u8)((_wValue) >> 8))
#define LowByte(_wValue) ((u8)((_wValue) & 0x00FF))
#define HighNib(_bValue) ((u8)((_bValue) >> 4))
#define LowNib(_bValue) ((u8)((_bValue) & 0x0F))

extern long iabs(long s);
extern long imin (long a, long b);
extern long imax (long a, long b);
extern void GaitSelect(void);
extern short SmoothControl (short CtrlMoveInp, short CtrlMoveOut, byte CtrlDivider);


//-----------------------------------------------------------------------------
// Define Global variables
//-----------------------------------------------------------------------------
//[TIMING]
extern unsigned long 	lTimerCnt;   //used in WTimer overflow. Will keep a 16 bit overflow so we have a 32 bit timer 

// [Which Processor]
extern u8	g_bWhichBap;			// may need to know at run time if running on BAP40 vs Arc32...



// debug handler...
extern bool g_fDBGHandleError;


#ifdef __cplusplus
}
#endif
#endif
