//==============================================================================
// IOPINS - Handle some of the IO functions of the BAP - This is a merge of 
// 		some of definitions/functions from Minionbuilder (bap.h/c) and my
// 		work to integrate it with my own stuff and includes my speed up code
//
// Version 0.1 - Work in Progress
//
// Major Contributers: Kurt Eckhardt 
// 		Robert V Mahoney Jr
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//==============================================================================
#ifndef _IOPINS_H_
#define _IOPINS_H_

#include "baplib.h"		// make sure our family and speed is set!
#ifdef __cplusplus
extern "C" {
#endif

//==============================================================================
// Type Definitions and Macros
//==============================================================================

// Define some constants for our IO pins
enum {
	P0=0, P1, P2, P3, P4, P5, P6, P7, P8, P9, P10, P11, P12, P13, P14, P15,
	P16, P17, P18, P19, P20, P21, P22, P23, P24, P25, P26, P27, P28, P29, P30, P31, 
	_S_IN_, _S_OUT_,_S_HSERIAL_=(u8)-1, _S_HSERIAL2_=(u8)-2};

// OK lets try to define the C version of the Port Table...
typedef struct {
		volatile u8					*pPDR;			// The PDR of the pin.
		u8							bMask;			// The mask in the IO register
		u8							bPU_PM;			// Pin have PUCR register(high nibble) PMDR(low nibble)
} PORTITEM;

#define PI_LOW(pi)		(*(pi.pPDR) &= ~(pi.bMask))
#define PI_HIGH(pi) 	(*(pi.pPDR) |= pi.bMask)
#define PI_TOGGLE(pi)	(*(pi.pPDR) ^= pi.bMask)
#define PI_ISHIGH(pi)	(*(pi.pPDR) & pi.bMask)

// Note these do not handle pin mode register changes or pull-up registers.  It is assummed that
// this was handled earlier, but are fast switching from input to output...
// Here I am passing pointers instead of structure...
__inline__ void PPI_INPUT(PORTITEM *ppi) __attribute__ ((always_inline));
__inline__ void PPI_INPUT(PORTITEM *ppi)
{
	extern u8 bap_bpcrshadowregs [];
	u8 bIOPortIndex = (ppi->pPDR - &IO.PDR1.BYTE);
	volatile unsigned char *pPCR = &IO.PCR1 + bIOPortIndex;
	bap_bpcrshadowregs[bIOPortIndex] &= ~(ppi->bMask);				// Update shadow mask to say that the pin is now an input pin
	*pPCR = bap_bpcrshadowregs[bIOPortIndex];
}

__inline__ void PPI_OUTPUT(PORTITEM *ppi) __attribute__ ((always_inline));
__inline__ void PPI_OUTPUT(PORTITEM *ppi)
{
	// Next lets update PDR/PCR...
	extern u8 bap_bpcrshadowregs [];
	u8 bIOPortIndex = (ppi->pPDR - &IO.PDR1.BYTE);
	volatile unsigned char *pPCR = &IO.PCR1 + bIOPortIndex;
	bap_bpcrshadowregs[bIOPortIndex] |= ppi->bMask;				// Ok update shadow mask to say it is output
	*pPCR = bap_bpcrshadowregs[bIOPortIndex];
}


// Add a few defines for Arduino compatibility
#define HIGH 1
#define LOW 0
#define INPUT	0
#define OUTPUT  1	// not sure which way these are defined not sure it matters...
#define digitalWrite(pin, state) setpin(pin, state)
#define digitalRead(pin) in(pin)
// Notes on bPU_PM high nibble is the offset from PUCR1(-1) (0-> no Pull up register)
#define PUR_1			0x10						// uses PUDR1
#define PUR_5			0x20						// uses PUDR5

// Likewise low nibble is offset from PMCR1(-1)  Assume that same mask will work!
#define PMR_1			0x1							// uses PMR1
#define PMR_5			0x2							// uses PMR5

//==============================================================================
// Function Prototypes
//==============================================================================

extern void input (u8 _bpin, bool fPullup=false);
extern void output (u8 _bpin);
extern void	pinMode (u8 _bpin, u8 _bmode);	// some compatibility with Arduino command set.
extern void low (u8 _bpin);
extern void high (u8 _bpin);
extern void toggle (u8 _bpin);
extern void setpin (u8 _bpin, u8 _bstate);
extern u8 in(u8 _bpin);


// These should maybe go to their own file?
extern void pauseus (unsigned long _lpause);
extern void pause (unsigned long  _lpause);
extern void Sound (u8 bPin, unsigned long wDur, unsigned long wFreq);			// dur in mS and Freq in Herz.  or should it be like basic???
#define delay(time)	pause(time)
extern void MSound(u8 bPin, u8 cNotes, ...);

//==============================================================================
// Global variable definitions.
//==============================================================================

extern PORTITEM	(*g_pPortTable)[];				// used to map logical pins to IOPort and mask
extern byte	g_cPortItems;								// Count of port Items;

#ifdef __cplusplus
}
#endif
#endif /* _IOPINS_H_ */

