//==============================================================================
// Kurts - Extras - for the phoenix on an Arc32
//==============================================================================
#include "globals.h"
#include <stdio.h>
#include "..\baplib\hsio.h"
#include "..\baplib\diyxbee.h"
#include "..\baplib\servo.h"
#include "..\baplib\eeprom.h"

//==============================================================================
// Defines and Variables
//==============================================================================

u8			_TM_szIn[80];			// input prompt
u8			_TM_iszIn;	

u8			_TM_bDumpMode;			// What mode are we in.  Simple stuff for now

u16 		_TM_wDumpStart;			// what address to start at
u16			_TM_wDumpCnt;			// how many bytes to dump


u16 		g_wDebugLevel;


// Forward references
extern void TerminalMonitor (u8 fShowTerminalPrompt);
extern void	_TM_ProcessEEPromDumpCmd(void);
extern void _TM_ProcessServoOffsetsCmd(void);
extern void	_TM_ProcessDownloadSeqCmd(void);
extern void	_TM_ProcessViewSequenceCmd(void);
extern u16 ExtractNum(void);
extern void DumpEEProm(u16 wDumpStart, u16 wDumpCnt);

extern void ReadServoOffsets(void);
extern void	MoveAllServos(void);
extern void MoveAServo(u8 iServo, u16 wNewPos, u16 wTime);
extern void	ShowWhichServo(u8 iServo);
extern u8 WriteServoOffsets(void); 


//==============================================================================
// [TerminalMonitorInit] Init the terminal monitor
//==============================================================================
void TerminalMonitorInit()
{
	// setup initial values for where to start dumping from and count
	_TM_wDumpStart = 0;
	_TM_wDumpCnt = 32;			// dump 

	TerminalMonitor(1);
}

//==============================================================================
// [TerminalMonitor]Terminal monitor - Since on the Arc32 S_IN/S_OUT are on a hardware
// serial port, there is almost no overhead for me to monitor it for inputs.  as
// such might as well add in some stuff from external programs I have.  For now
// will add Sequence Downloader, EEPROM Dump, and set debug on and off...
//==============================================================================

void TerminalMonitor (u8 fShowTerminalPrompt)
{
	u8 cbRead;
	if (fShowTerminalPrompt) {
_TM_ShowPrompt:

		DBGPrintf("\n\nPhoenix-Arc32 keyboard monitor\n\r");
		DBGPrintf("  D [<start addr>][<cnt>] - Dump EEPROM memory\n\r");
		DBGPrintf("  O - Enter Servo Offset mode\n\r");
		DBGPrintf("  S - Download sequence(s) - VB call only\n\r");
		DBGPrintf("  T <n>] - Set or show debug Trace level\n\r");
		DBGPrintf("  V <n> - View Sequence\n\r");
		DBGPrintf(" : ");
		fShowTerminalPrompt = 0;
	}
	
	// See if there is any keyboard input available.
	// Question: If I have some input should I wait until I get a CR?  Will for now
	if (hserstat(1,HSERSTAT_INPUT_AVAILABLE)) {
		// we have some input available...
		cbRead = hserin (1, 0xffffffffL, _TM_szIn, sizeof(_TM_szIn), 13);
		if (_TM_szIn[0] == 13)
			goto _TM_ShowPrompt;			// I hate Gotos...

		else if ((_TM_szIn[0] == 't') || (_TM_szIn[0] == 'T')) {	// The user wants to see or set the debug level
			if (_TM_szIn[1] != 13) {
				_TM_iszIn = 1;
				g_wDebugLevel = ExtractNum();
			}
			DBGPrintf("\n\rCurrent Debug Level: %d(%x)\n\r", g_wDebugLevel, g_wDebugLevel);
		} 
		else if ((_TM_szIn[0] == 'd') || (_TM_szIn[0] == 'D'))	// The user wants to Display EEPROM memory
			_TM_ProcessEEPromDumpCmd();
			
		else if ((_TM_szIn[0] == 'o') || (_TM_szIn[0] == 'O'))	// The user wants to update the servos offsets
			_TM_ProcessServoOffsetsCmd();

		else if ((_TM_szIn[0] == 's') || (_TM_szIn[0] == 'S'))	// The user wants to download a sequence or sequences...
			_TM_ProcessDownloadSeqCmd();
		else if ((_TM_szIn[0] == 'v') || (_TM_szIn[0] == 'V'))	// The user wants to view a sequence.
			_TM_ProcessViewSequenceCmd();

		DBGPrintf(": ");
	}		
}

//==============================================================================
//==============================================================================
// EEPROM - Dump code...
//==============================================================================
//0 - standard hex byte dump (h)
//1 - unsigned Decimal - words (w)
//2 - signed decimal words (s)
//------------------------------------------------------------------------------
void _TM_ProcessEEPromDumpCmd()
{
	if (_TM_szIn[1] == 13)
		DumpEEProm(_TM_wDumpStart, _TM_wDumpCnt);
	else if ((_TM_szIn[1] == 'h') || (_TM_szIn[1] == 'H'))
		_TM_bDumpMode = 0;
		
	else if ((_TM_szIn[1] == 'w') || (_TM_szIn[1] == 'w'))
		_TM_bDumpMode = 1;
		
	else if ((_TM_szIn[1] == 's') || (_TM_szIn[1] == 'S'))
		_TM_bDumpMode = 2;
	else {
		_TM_iszIn = 1;
		_TM_wDumpStart = ExtractNum();
		if (_TM_szIn[_TM_iszIn] == '=') {
			// we are going to try to update the EEPROM!
			_TM_iszIn = _TM_iszIn + 1;
			
			while (_TM_szIn[_TM_iszIn] != 13) {	// let walk through the items
				_TM_wDumpCnt = ExtractNum(); 	// reuse this variable.
				
				// This pass of the code I will only update in bytes or words, but not signed
				if (_TM_bDumpMode == 0) {
					//Writedm _TM_wDumpStart,[_TM_wdumpCnt.lowbyte]
					_TM_wDumpStart = _TM_wDumpStart + 1;
				} else {
					//Writedm _TM_wDumpStart,[_TM_wdumpCnt.highbyte, _TM_wdumpCnt.lowbyte] // Write 2 bytes
					_TM_wDumpStart = _TM_wDumpStart + 2;
				}
			}
		} else {
			if (_TM_szIn[_TM_iszIn] == ' ') {
				// assume that we have a count specified here
				_TM_iszIn = _TM_iszIn + 1;
				_TM_wDumpCnt = ExtractNum();
			}
			
			DumpEEProm(_TM_wDumpStart, _TM_wDumpCnt);
		}
	}

	_TM_wDumpStart = _TM_wDumpStart + _TM_wDumpCnt; // assume we will start after the last write...
}
	
							
//------------------------------------------------------------------------------
// Extract Number from input line
//------------------------------------------------------------------------------

u16 ExtractNum(void)
{
	u16 	wNum;	
	u8 		fContinue;
	
	wNum = 0;
	fContinue = 1;
	// skip any leading blanks...
	while (_TM_szIn[_TM_iszIn] == ' ')
		_TM_iszIn++;

	// lets parse the strings to see if we have a start address or not...
	if ((_TM_szIn[_TM_iszIn] == '0') && ((_TM_szIn[_TM_iszIn+1] == 'x') || (_TM_szIn[_TM_iszIn+1] == 'X'))) {
		// hex number 
		_TM_iszIn = _TM_iszIn+2;
		do { 
			if ((_TM_szIn[_TM_iszIn] >= '0') && (_TM_szIn[_TM_iszIn] <= '9')) {
				wNum = wNum * 16 + _TM_szIn[_TM_iszIn]-'0';
				_TM_iszIn++;
			}	
			else if ((_TM_szIn[_TM_iszIn] >= 'a') && (_TM_szIn[_TM_iszIn] <= 'f')) {
				wNum = wNum * 16 + _TM_szIn[_TM_iszIn]-'a'+10;
				_TM_iszIn++;
			}
			else if ((_TM_szIn[_TM_iszIn] >= 'A') && (_TM_szIn[_TM_iszIn] <= 'F')) {
				wNum = wNum * 16 + _TM_szIn[_TM_iszIn]-'A'+10;
				_TM_iszIn++;
			}	
			else 
				fContinue = 0;
		} while(fContinue);
	} else {
		// assume decimal number
		do { 
			if ((_TM_szIn[_TM_iszIn] >= '0') && (_TM_szIn[_TM_iszIn] <= '9')) {
				wNum = wNum * 10 + _TM_szIn[_TM_iszIn]-'0';
				_TM_iszIn++;
			}
			else 
				fContinue = 0;
		} while(fContinue);
	}

	return wNum;
}	

//------------------------------------------------------------------------------
// DumpEEProm:
//------------------------------------------------------------------------------
void DumpEEProm(u16 wDumpStart, u16 wDumpCnt)
{
	u8	abDumpMem[16];
	u8 iDump;
	u16 wb;

	DBGPrintf("Start Dump: %x  Count: %d\n\r", wDumpStart, wDumpCnt);
	while (wDumpCnt > 0) {

		// Pass in the default address for the built in eeprom
		EEPROMRead (0xA0, wDumpStart, min (wDumpCnt, 16), abDumpMem);
		
		DBGPrintf ("%04x:", wDumpStart);
		
		// first dump the characters out in hex
		for (iDump = 0; iDump < 16; iDump++) { 
			if (_TM_bDumpMode == 0) {
				if (iDump < wDumpCnt)
					DBGPrintf ("%02x ", abDumpMem[iDump]);
				else
					DBGPrintf("   ");
			}
			else if ((iDump & 0x1) == 0){	// sort-of crap - but word size so only output every other byte
				if (iDump < wDumpCnt) {
					if (_TM_bDumpMode == 1) {
						wb = (abDumpMem[iDump] << 8) | abDumpMem[iDump+1];
						DBGPrintf("%6u ", wb);
					} else {
						wb = (abDumpMem[iDump] << 8) | abDumpMem[iDump+1];
						DBGPrintf("%6s ", wb);	// will check sign itself...
					}
				} else
					DBGPrintf("       ");
			}
		}
		
		// then dump them out in ascii
		for (iDump = 0; (iDump < 16) && (iDump < wDumpCnt); iDump++) {
			if ((abDumpMem[iDump] > 0x1f) && (abDumpMem[iDump] < 0x7f))
				DBGPrintf("%c", abDumpMem[iDump]);
			else
				DBGPrintf(".");
		}
		
		DBGPrintf("\n\r");	// output the cr to end the line

		// setup for the next line if any
		if (wDumpCnt > 16) {
			wDumpCnt -=16;
			wDumpStart += 16;
		} else
			wDumpCnt = 0;
	}
}	

//------------------------------------------------------------------------------
// View Sequences command processing
//------------------------------------------------------------------------------
void _TM_ProcessViewSequenceCmd(void)
{
	u16 wGPSeqPtr;
	u8	bGPCntServos;
	u8 	bGPCntSteps;
	u8 	bGPServoNum;
	u16	wB;
	u8	ab[100];
	u16 wGPStepTime;
	u8 bGPStepNum;
	u8	bPin;
		
	if (_TM_szIn[1] == 13)
		_TM_wDumpStart = 0;	// assume the first one.
	else {
		_TM_iszIn = 1;
		_TM_wDumpStart = ExtractNum();
	}
	
	// Now lets try to walk through the memory to dump out the data about the selected sequence.
	EEPROMRead (0xA0, ARC32_SSC_OFFSET + _TM_wDumpStart*2, 2, (u8*)&wGPSeqPtr);
	
	// make sure there is a sequence...
	if ((wGPSeqPtr != 0)  && (wGPSeqPtr != 0xffff)) {
		// now lets get the header information, will skip sequence number
		EEPROMRead (0xA0, wGPSeqPtr+1, 1, &bGPCntServos);  // bugbug:: could read in array with both...
		EEPROMRead (0xA0, wGPSeqPtr+2, 1, &bGPCntSteps);

		// now lets read in the pins
		wGPSeqPtr += 3;		 // point to start of pin information.
		DBGPrintf("Sequence: %u Cnt Servos: %u Steps: %u\n\rServos:        ", 	_TM_wDumpStart, bGPCntServos, bGPCntSteps);

		for (bGPServoNum = 0;  bGPServoNum < bGPCntServos; bGPServoNum++) {
			EEPROMRead(0xa0, wGPSeqPtr, 1, &bPin);
			DBGPrintf("%4u ",bPin);
			wGPSeqPtr += 3;
		}
		DBGPrintf("\n\r");

		wGPSeqPtr += 2; // skip over the N-1 time
		// now walk through the steps
		for (bGPStepNum = 0; bGPStepNum < bGPCntSteps; bGPStepNum++) {
			EEPROMRead(0xa0, wGPSeqPtr, 2*bGPCntServos+2, ab);
			wGPStepTime = (ab[2*bGPCntServos] << 8) | ab[2*bGPCntServos+1];
			DBGPrintf("S: %3u %4u ", bGPStepNum, wGPStepTime);

			for (bGPServoNum = 0; bGPServoNum < bGPCntServos; bGPServoNum++) {
				// no longer converting the stored pulses to hservo so don't need any translation
				wB = ( ab[bGPServoNum*2] << 8) + ab[bGPServoNum*2+1];
				DBGPrintf("%4u ", wB);
			}
			DBGPrintf("\n\r");
			wGPSeqPtr += 2*bGPCntServos + 2;	// increment to the next sequence
		}
	}
}
					


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@	
//==================================================================================================
// Servo Offsets sub-section
//==================================================================================================
//------------------------------------
// Hex robot - Servo zero point finder
// 
// This version is written for the Atom Pro
// *** without an SSC-32 *** and uses HSERVO
// 
// It will store the values in the EEPROM of the BAP.
// Will emulate the SSC-32 in that we store the values out at memory
// locations 32-63.  However I will also store out a checksum at
// memory location 31, which I will compare against on the read to
// make sure the information looks valid.  
// Since we want a range of at least +- 8 degrees 

#define NUMSERVOS	18
//[SSC PIN NUMBERS]
// Actual pin number will now be in CFG file...
u8	g_ServoTable[] = {cRRCoxaPin,cRRFemurPin,cRRTibiaPin,
							cRMCoxaPin,cRMFemurPin,cRMTibiaPin,
							cRFCoxaPin,cRFFemurPin,cRFTibiaPin,
							cLRCoxaPin,cLRFemurPin,cLRTibiaPin,
							cLMCoxaPin,cLMFemurPin,cLMTibiaPin, 
							cLFCoxaPin,cLFFemurPin,cLFTibiaPin};
							
u8		_TT1[] = {"RR Hip Hor"};
u8		_TT2[]	= {"RR Hip Vert"};
u8		_TT3[] = {"RR Knee"}; 
u8		_TT4[] = {"MR Hip Hor"};
u8		_TT5[] = {"MR Hip Vert"};
u8		_TT6[] = {"MR Knee"};
u8		_TT7[] = {"FR Hip Hor"}; 
u8		_TT8[] = {"FR Hip Vert"};
u8		_TT9[] = {"FR Knee"};
u8		_TT10[] = {"RL Hip Hor"};
u8		_TT11[] = {"RL Hip Vert"};
u8		_TT12[] = {"RL Knee"};
u8		_TT13[] = {"ML Hip Hor"};
u8		_TT14[] = {"ML Hip Vert"};
u8		_TT15[] = {"ML Knee"};
u8		_TT16[] = {"FL Hip Hor"};
u8		_TT17[] = {"FL Hip Vert"};
u8		_TT18[] = {"FL Knee"};
u8	*g_ServoNameTable[] = {_TT1, _TT2, _TT3, _TT4, _TT5, _TT6, _TT7, _TT8, _TT9, _TT10,
						   _TT11, _TT12, _TT13, _TT14, _TT15, _TT16, _TT17, _TT18};

// We will keep a set of offsets  
#define MAXOFFSET	2000					// what is the maximum offset we will allow
#define SERVOSAVECNT	32				
short g_asServoOffsets[SERVOSAVECNT];			// This is up in phoenix_v32.bas..

//==============================================================================
// Complete initialization
//==============================================================================
void _TM_ProcessServoOffsetsCmd() 
{



	u8 bChar;
	u8	fChanged;
	u8	cChanges;
	u8	fSSCReboot;
	u8	bPlusMinusOffset;		// How much to offset when the + or - is hit

	u8	iServoTable;			// Which servo table entry is current
	u8	bCurServo;				// Get the actual servo number to keep from having to double table lookup
	u8	b;

	fSSCReboot = 1;				// process this the first time through...
	fChanged = 0;
	cChanges = 0;
	bPlusMinusOffset = 0;
	iServoTable = 0;
	bCurServo = 0;		// Several of these inits not needed but got rid of some warnings...
		
	for(;;) {
		if (fSSCReboot) {
			fSSCReboot = 0;
			iServoTable = 0;
			bCurServo = g_ServoTable[iServoTable];
			for (b = 0; b < sizeof(g_asServoOffsets)/sizeof(g_asServoOffsets[0]); b++)
				g_asServoOffsets[b] = 0; 
			pause(500);	// give some time to setup
			
			ReadServoOffsets();
		
			// Now lets display the offsets for the different servos
			for (b = 0; b <NUMSERVOS; b++) {
				bCurServo = g_ServoTable[b];
				DBGPrintf ("%c %s = %d\n\r", 'A'+b, g_ServoNameTable[b], g_asServoOffsets[bCurServo]);
			}
		
			pause(500);
			
			MoveAllServos();
		
			DBGPrintf("Use keyboard to adjust the servos\n\r");
			DBGPrintf(" +-  - to increment/decrement the current servo\n\r");
			DBGPrintf(" 0   - Will zero out that offset");
			DBGPrintf(" 1-9 - Set the increment default 50 Num * 10\n\r");
			DBGPrintf(" @   - Clear all offsets from SSC\n\r");
			DBGPrintf(" *   - to choose next servo or a-z to choose specific...\n\r");
			DBGPrintf(" <$> - Return to main program\n\r");
			DBGPrintf(" <entr> - to save away the updated values\n\r");
		
			iServoTable = 0;
			bCurServo = g_ServoTable[iServoTable];
			ShowWhichServo(iServoTable);
			fChanged = 0;
			fSSCReboot = 0;
			bPlusMinusOffset = 50;
			cChanges = 0;	// so we don't keep going and going on the keyapd...
		}		
		//==============================================================================
		// Main Loop
		//==============================================================================
		
		// warning serin only works when you are in it and the character(s) arrive.  So try
		// to be in the majority of the time.  For this case where we are simply doing
		// servo find center it will probably not be that much of an issue.
		hserin (1, 0xffffffffL, &bChar, 1, 0xffff);
	
		if (bChar == '$')	// Escape... Return to main loop
			return;
		else if (bChar == '-') {
			if (g_asServoOffsets[bCurServo] > -MAXOFFSET) {
				// Make sure we did not move it out of range...
				if (g_asServoOffsets[bCurServo] < (-MAXOFFSET + bPlusMinusOffset))
					g_asServoOffsets[bCurServo] = -MAXOFFSET;
				else
					g_asServoOffsets[bCurServo] -= bPlusMinusOffset;
	
				MoveAServo(bCurServo, g_asServoOffsets[bCurServo], 128);	// move each servo to center point
				cChanges++;
				if (cChanges > 5){
					DBGPrintf("\n\r : %s\n\r", g_asServoOffsets[bCurServo]);
					cChanges = 0;
				}
	
				fChanged = 1;
			}	
   			else
    			MSound(SOUND_PIN, 1, 150, 1750);
		}  
		else if (bChar == '+') { 
			if (g_asServoOffsets[bCurServo] < MAXOFFSET) {
				// Make sure we did not move it out of range...
				if (g_asServoOffsets[bCurServo] > (MAXOFFSET - bPlusMinusOffset))
					g_asServoOffsets[bCurServo] = MAXOFFSET;
				else
					g_asServoOffsets[bCurServo] += bPlusMinusOffset;
				
				MoveAServo(bCurServo, g_asServoOffsets[bCurServo], 128);	// move each servo to center point
				fChanged = 1;
				cChanges++;
				if (cChanges > 5) {
					DBGPrintf("\n\r : %s\n\r", g_asServoOffsets[bCurServo]);
					cChanges = 0;
				}
			}
   			else
    			MSound(SOUND_PIN, 1, 150, 1750);
		} 		
 		else if (bChar == '0') {
			g_asServoOffsets[bCurServo] = 0;		// clear it back to logical zero
			MoveAServo(bCurServo, g_asServoOffsets[bCurServo], 128);	// move each servo to center point
			fChanged = 1;
		}
		else if ((bChar >= '1') && (bChar <= '9')) 
			bPlusMinusOffset = (bChar - '0') * 10;	// set the increment
			
 		else if ((bChar == '*') || (bChar >= 'a' && bChar <= ('a'+NUMSERVOS-1)) || (bChar >= 'A' && bChar <= ('A'+NUMSERVOS-1))) {
    		if (fChanged) {
				DBGPrintf("\n\r : %s\n\r", g_asServoOffsets[bCurServo]);
	 			fChanged = 0;
				cChanges = 0;
			}
	 		if (bChar == '*') {
	  			iServoTable = iServoTable + 1;
	  			if (iServoTable == NUMSERVOS)
					iServoTable = 0;
			}
			else if (bChar >= 'a' && bChar <= ('a'+NUMSERVOS-1)) 
				iServoTable = bChar - 'a';
			else
				iServoTable = bChar - 'A';
			bCurServo = g_ServoTable[iServoTable];
  			ShowWhichServo(iServoTable);
  		}	
		else if (bChar == '@') { 				// Clear all of the values from the SSC
			for (b = 0; b < sizeof(g_asServoOffsets)/sizeof(g_asServoOffsets[0]); b++)
				g_asServoOffsets[b] = 0; 
   			MSound(SOUND_PIN, 2, 75, 1600, 50, 3300);
			fSSCReboot = WriteServoOffsets(); // ok lets write out the updated values...
		}
			
		else if (bChar && (bChar <= 13)) { 			// handle cr or lf or...
   			MSound(SOUND_PIN, 2, 75, 1600, 50, 3300);
			fSSCReboot = WriteServoOffsets(); // ok lets write out the updated values...
  		}
	
		if (fSSCReboot) {
			DBGPrintf("+++ Reset after values updated +++\n\r");
		}
	}	
}

//==============================================================================
// subroutine: MoveAServo
// Move a specific servo to the specified location and wait for it to get there
//==============================================================================
void MoveAServo(u8 iServo, u16 wNewPos, u16 wTime)
{

	HServoGroupMoveStart();		
	HServo(iServo, wNewPos);				// BUGBUG: What coordiate system am I in???
	HServoGroupMoveCommit(wTime);
	HServoWait(1L<<iServo);
}
	
//==============================================================================
// subroutine: MoveAllServos
// Calls Hservo to move all of the servos to their current zero location
//==============================================================================
void MoveAllServos()
{
	u8 iServoTable;
	u8 bCurServo;
	for (iServoTable = 0; iServoTable < NUMSERVOS; iServoTable++) {
		bCurServo = g_ServoTable[iServoTable];
		MoveAServo(bCurServo, 1500 + g_asServoOffsets[bCurServo], 0);	// move each servo to center point
	}
}

//==============================================================================
// Subroutine: ShowWhichServo
// Helps to let the user know which servo they are adjusting
// then it puts it to the current position...
//==============================================================================
void ShowWhichServo(u8 iServo)
{
	MoveAServo(g_ServoTable[iServo], 1100, 128);	// move each servo to center point
	pause (128);
	MoveAServo(g_ServoTable[iServo], 1900, 128);	// move each servo to center point
	pause (128);
	// move to current position which is 1500 + our offset...
	MoveAServo(g_ServoTable[iServo], 1500+ g_asServoOffsets[g_ServoTable[iServo]], 128);	// move each servo to center point
	pause (128);

	DBGPrintf ("%c %s = %d\n\r", 'A'+iServo, g_ServoNameTable[iServo], g_asServoOffsets[g_ServoTable[iServo]]);
}

//==============================================================================
// Subroutine: ReadServoOffsets
// Will read in the zero points that wer last saved for the different servos
// that are part of this robot.  
//
//==============================================================================
void ReadServoOffsets(void)
{
	u8 bCSIn;
	u8 bCSCalc;
	u8 Index; 
		
	EEPROMRead(0xA0, 31, 1, &bCSIn);
	EEPROMRead(0xA0, 32, 32, (u8*)g_asServoOffsets);
	EEPROMRead(0xA0, 64, 32, (u8*)(&g_asServoOffsets[16]));
	//readdm 31, [ bCSIn]
	//readdm 32, [str aServoOffsets\32]	; We are storing words now.
	//readdm 64, [str aServoOffsets(16)\32]
	bCSCalc = 0;
	for (Index = 0; Index < SERVOSAVECNT; Index++)
		bCSCalc += (g_asServoOffsets[Index] & 0xff) + ((g_asServoOffsets[Index] >> 8) & 0xff);
		
	if (bCSCalc != bCSIn) {
		MSound(SOUND_PIN, 3, 60, 2000, 100, 2500, 60, 2000);	// make some noise...
		DBGPrintf("--- Invalid Servo Offsets ---\n\r");
		DBGPrintf(" CS in: %x CS calc: %x\n\r", bCSIn, bCSCalc);
		for (Index = 0; Index < sizeof(g_asServoOffsets)/sizeof(g_asServoOffsets[0]); Index++)
			g_asServoOffsets[Index] = 0; 
	}

}
// Now up in the main file...
	
//==============================================================================
// Subroutine: WriteServoOffsets
// Will write out the current servo offsets to the EEPROM on the BAP29  
//
//==============================================================================
u8 WriteServoOffsets()
{

	// OK First calculate the checksum
	// We will do something simple like add all of the bytes to each other.
	// bugbug:: If like bap28 can not write more than 32 bytes at a time so
	// will need to break up into smaller pieces for larger writes like
	// offset for 18 servos.
	u8 bCSCalc = 0;
	u8 i;
	
	DBGPrintf ("+++ Write Servo OffsetsOffsets +++\n\r");

	for (i = 0; i < SERVOSAVECNT; i++)
		bCSCalc += (g_asServoOffsets[i] & 0xff) + ((g_asServoOffsets[i]>>8) & 0xff);
	
	// Now write out checksum, followed by offset data
//	writedm 31, [ bCSCalc]
	EEPROMWrite(0xA0, 31, 1, &bCSCalc);
	pause (10);
	//writedm can only handle 32 bytes a time so will split to two writes
	EEPROMWrite(0xA0, 32, 32, (u8*)g_asServoOffsets);
	//writedm 32, [str g_asServoOffsets\32]		// bugbug should check to make sure that count
	pause (10);
	EEPROMWrite(0xA0, 64, 32, (u8*)(&g_asServoOffsets[16]));
	pause (10);
	
	return 1; // always jump to the start after this...
}

//##################################################################################################
//------------------------------------------------------------------------------
// Download sequences.
//------------------------------------------------------------------------------ 

//==============================================================================
// [_TM_ProcessDownloadSeqCmd]
//==============================================================================
typedef struct {
	u8 		cSeqs;
	u16		cbTotal;
	u8		iSeqDl;
	u8		abHdr[6];
	u8		iSeq;
	u16		cbSeq;
	u16		wSeqStart;
	u16		wBuffStart;
	u16		cbLeft;
	u16		wBufferStart;
	u8		cbSSCWrite1stBlock;
	u8		ab[70];		// Room for a couple of read in 32 bytes at a time, also leave room for ack or the like...
} TMPDLSD;				// data to pass back and forth to write function...
extern void ClearSeqPtrs(void);	// clear the sequence pointers
extern void WriteBlockToSSC(u8 bState, TMPDLSD *pdlsd);



void _TM_ProcessDownloadSeqCmd(void)
{
	u8		i;
	u8		bAck;
	u8		bChksum;
	u8		bDLReadState;
	TMPDLSD dlsd;		// This has all of the cownload sequence information we will pass back and forth...

	// The VB app will pass us an ascii string with the number of sequences and cbTotal as ascii.
	// need to extract that from the command line...
	_TM_iszIn = 1;
	dlsd.cSeqs = ExtractNum(); 		// extract the count of sequences
	dlsd.cbTotal = ExtractNum(); 	// likewise the expected total number of bytes

	// do a little validation to make sure the user did not enter this by accident...
	// need at least one sequence
	if ((dlsd.cSeqs == 0) || (dlsd.cSeqs > 20) || (dlsd.cbTotal < (u16)(dlsd.cSeqs*12)) || (dlsd.cbTotal > 16383)) {
		// either no steps or too many or not enough bytes for even 1 servo and step per sequence
		// or obviously won't fit in EEPROM, then probably garabage, so bail before we do anything
		DBGPrintf("*** Download Command Error ***\n\r");
		return;
	}
	// Calculate return number that the other will understand as an ack
	// should add some more validation here
	bChksum = dlsd.cSeqs + (dlsd.cbTotal >> 8) + (dlsd.cbTotal & 0xff);
	bAck = !bChksum;
	hserout	(1, &bAck, 1);		// send an Ack back to the caller.
	ClearSeqPtrs();				// clear the sequence pointers

	// now we need to loop through all of the sequences that are supposed to be downloaded
	for (dlsd.iSeqDl = 0; dlsd.iSeqDl < dlsd.cSeqs; dlsd.iSeqDl++) {
		// We should now get a header for the next sequnce
		// BUGBUG should probably put a timeout in here...
		hserin (1, (u32)-1,  dlsd.abHdr, 6, 0xffff);		// read in 5 byte header + 
		bChksum = dlsd.abHdr[0]+dlsd.abHdr[1]+dlsd.abHdr[2]+dlsd.abHdr[3]+dlsd.abHdr[4];
	
		// we have hopefully a valid header...
		bAck = !bChksum;		// invert what we received to send back		
		hserout	(1, &bAck, 1);		// send an Ack back to the caller.

		dlsd.iSeq = dlsd.abHdr[0];

		dlsd.wSeqStart = (dlsd.abHdr[1] << 8) + dlsd.abHdr[2];	// Where it was in EEPROM file. 
		dlsd.cbSeq = (dlsd.abHdr[3] << 8) + dlsd.abHdr[4];		// This sequence size 

		dlsd.cbLeft = dlsd.cbSeq;
		dlsd.wBuffStart = 0;
		bDLReadState = 0;

		while (dlsd.cbLeft >= 32) {
			// also put in timeout...
			hserin (1, (u32)-1, &dlsd.ab[dlsd.wBuffStart], 32+1, 0xffff);	// Read in next 32 bytes + checksum
			
			bChksum = 0;
			for (i = 0; i < 32; i++)
				bChksum += dlsd.ab[dlsd.wBuffStart+i];

			if (bChksum != dlsd.ab[dlsd.wBuffStart+32])
				goto _TMDL_Error;

			WriteBlockToSSC(bDLReadState, &dlsd);			// Do our call off
			bDLReadState = 1;						// not our first call						

			bAck = !bChksum;		// invert what we received to send back		
			hserout	(1, &bAck, 1);		// send an Ack back to the caller.
			dlsd.cbLeft -= 32;
		}
		
		if (dlsd.cbLeft) {
			hserin (1, (u32)-1, &dlsd.ab[dlsd.wBuffStart], dlsd.cbLeft+1, 0xffff);	// Read in next 32 bytes + checksum
			bChksum = 0;
			for (i = 0; i < dlsd.cbLeft; i++)
				bChksum += dlsd.ab[dlsd.wBuffStart+i];

			if (bChksum != dlsd.ab[dlsd.wBuffStart+dlsd.cbLeft])
				goto _TMDL_Error;
			
			// now lets finish writing out the sequence.
			dlsd.cbLeft += dlsd.wBuffStart;		// this should be the total number of bytes left to output
			if (dlsd.cbLeft >= 32) {
				WriteBlockToSSC(1, &dlsd);		// Do a full write...
				dlsd.cbLeft -= 32;
			}
			
			bAck = !bChksum;		// invert what we received to send back		
			hserout	(1, &bAck, 1);		// send an Ack back to the caller.
		
		}
		
		// After we finish receiving a sequence we can now process it.  At the end of that echo one last Ack to
		// let the other side know that we completed it...
		WriteBlockToSSC(2, &dlsd);			// Do possibly a partial write and any cleanup that may be necessary.
		hserout	(1, &bAck, 1);		// send an Ack back to the caller.
	}
	return;

_TMDL_Error:
	DBGPrintf("*** Sequence Download Failed: Checksum Error ***\n\r");
}	

	
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// 				ARC 32 version
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	
//===============================================================================
// ClearSeqPtrs
//===============================================================================
void ClearSeqPtrs(void)	// clear the sequence pointers
{
	int i;
	u8 b;
	b = 0;

	// doing it the slow way...
	for (i=0; i < 20; i++)
		EEPROMWrite(0xA0, ARC32_SSC_OFFSET+i, 1, &b);
}

//===============================================================================
// WriteBlockToSSC - Version 2 - Arc32 version
//		before we start writing out to the SSC, we will receive one block and write
// 		it.  The SSC is more efficient if it writes on 32 byte boundaries, so
//		we will try to align all of our writes.   This implies the first buffer
//		we will only write part of it out, 
//		bState:
// 			0 - First output - need to output pointer, plus part or read up to page boundary
// 			1 - Main state - output 32 bytes, copy rest down to start...
// 			2 - output final part...
//		
//===============================================================================
void WriteBlockToSSC(u8 bState, TMPDLSD *pdlsd)
{
	u8	i;
		
	// will need to write a max of 32 bytes at a time preferably starting at 32 bit boundry...
	// We will use the memory address that was passed us...

	// first output the address to the start the of SSC
	if (bState == 0) { 		// we received first buffer.
		// first write out pointer to the data for this sequence.  We will offset the 
		// address by ARC32_SSC_OFFSET to save room for other stuff 

		pdlsd->wSeqStart += ARC32_SSC_OFFSET;
		// first output the address to the start the of SSC - Note since this is all internal
		// we may as well offset the value and save it in our own number ordering...
		EEPROMWrite(0xA0, ARC32_SSC_OFFSET + pdlsd->iSeq*2, 2, (u8*)&pdlsd->wSeqStart);
		pause (10);

		// The SSC-32 likes to do writes starting at 32 byte boundaries.  So calculate how many of
		// our first 32 bytes we will output in this call.  This value will also be used then by the
		// receive function to know how far off in our buffer to offset to do the next read...
		pdlsd->cbSSCWrite1stBlock = 32 - (pdlsd->wSeqStart & 0x1f);	// this is the offset into the 

		// ok let write the first part out
		EEPROMWrite(0xA0, pdlsd->wSeqStart, pdlsd->cbSSCWrite1stBlock, pdlsd->ab);

		// now the gunk, to make the code easy I am going to copy the rest of the buffer down that we did not output yet...
		if (pdlsd->cbSSCWrite1stBlock != 32) {
			for (i = pdlsd->cbSSCWrite1stBlock; i< 32; i++)
				pdlsd->ab[i-pdlsd->cbSSCWrite1stBlock] = pdlsd->ab[i];	// copy the parts down from the end to the start
		}
	
		pdlsd->wSeqStart = (pdlsd->wSeqStart + 32) & 0xffe0;			// setup to the start of the next sequence

		pdlsd->wBuffStart = 32 - pdlsd->cbSSCWrite1stBlock;				// This is where in the buffer we should setup to do the next read...

	} else if (bState == 1) {	// then normal 32 byte outputs.

		EEPROMWrite(0xA0, pdlsd->wSeqStart, 32, pdlsd->ab);				// Output the block

		// now junk again copy down the part of our last read that was not output
		// On the secondary reads, we started the read at the buffer location of
		// ab(32-cbSSCWrite1stBlock) so we will have 32-cbSSCWrite1stBlock bytes that go into our
		// buffer after what we wrote...
		if (pdlsd->cbSSCWrite1stBlock != 32) {
			for (i = 0; i < 32-pdlsd->cbSSCWrite1stBlock; i++)
				pdlsd->ab[i] = pdlsd->ab[32+i];	// copy the parts down from the beginning of the next block...
		}
		pdlsd->wSeqStart += 32;
	} else {	// bstate = 2, do the last buffer...
		if (pdlsd->cbLeft) {
			pause (10);
			EEPROMWrite(0xA0, pdlsd->wSeqStart, pdlsd->cbLeft, pdlsd->ab);
		}
	}
}

